#!/usr/bin/runhaskell
{-# LANGUAGE FlexibleContexts #-}

import Prelude hiding (fail, uncurry)
import Tct.Configuration
import Tct.Interactive
import Tct.Instances
import qualified Tct.Processors as Processor
import qualified Termlib.Repl as TR

import qualified Data.Set as Set
import qualified Tct.Method.DP.DependencyGraph as DG

import Abbrevs
import qualified Raml as Raml


main :: IO ()
main = tct config

config :: Config
config = defaultConfig { strategies = strats
                       -- , interactiveShowProofs = True
                       , recompile = False } 
  where
    strats = 
      [ Raml.raml  ::: strategy "raml" (optional (maybeArg naturalArg) "inner-timeout" (Just 15)
                                        :+: optional boolArg "greedy" False
                                        :+: optional naturalArg "degree" (Nat 2)
                                        :+: optional directsArg "base" [PopStar, Polys, Matrices])
      , mxs        ::: strategy "matrices" (optional naturalArg "mindegree" (Nat 1)
                                            :+: optional naturalArg "maxdegree" (Nat 3)
                                            :+: optional boolArg "uargs" True)
      , pxs        ::: strategy "polys" (optional naturalArg "mindegree" (Nat 1)
                                         :+: optional naturalArg "maxdegree" (Nat 3)
                                         :+: optional boolArg "uargs" True)
      , sem        ::: strategy "semantics" (optional naturalArg "mindegree" (Nat 1)
                                             :+: optional naturalArg "maxdegree" (Nat 3)
                                             :+: optional boolArg "uargs" True)
      , simple     ::: strategy "simple" (optional boolArg "dp" True)
      , trivials   ::: strategy "trivials"
      -- ICC
      , simpleICC   ::: strategy "simpleICC" (optional boolArg "usePOP" True)
      , bspopstar   ::: strategy "bspopstar"
      , bspopstarPS ::: strategy "bspopstarPS"
      -- WST 13
      , toDP ::: strategy "withDP" ]

--------------------------------------------------------------------------------
-- conveniences                
--------------------------------------------------------------------------------

app m = apply m >> return ()
loadDPi s = loadIRC s >> app toDP

loadUncurried fp = 
  do load fp
     apply uncurry
     [p] <- problems
     modifyInitialWith $ const $ f p
    where f p = p { TR.startTerms = TR.BasicTerms ds cs 
                  , TR.strategy = TR.Innermost} 
            where rs = TR.allComponents p
                  ds = TR.definedSymbols rs
                  cs = TR.constructors rs
                

--------------------------------------------------------------------------------
-- combinators
--------------------------------------------------------------------------------


mxs (Nat min :+: Nat max :+: uargs) = fastest $ concat [ [mx i i `withUsableArgs` uargs
                                                         , mx (i + 1) i  `withUsableArgs` uargs] 
                                                         | i <- [min..max] ]
pxs (Nat min :+: Nat max :+: uargs) = fastest [ poly  `withUsableArgs` uargs `withDegree` Just i | i <- [min..max] ]
sem args = mxs args  `orFaster` pxs args

----------------------------------------------------------------------
-- transformations
--------------------------------------------------------------------------------

wgOnUsable = wg 2 1 `wgOn` WgOnTrs

removeLeafs 0 = exhaustively $ whenNonTrivial $ removeLeaf (mx 1 0)
removeLeafs i = removeLeaf (spopstarPS `withDegree` Just i) 
                <> (removeLeaf (mx i i) 
                    <||> removeLeaf (mx (i + 1) i)
                    <||> when (i == 1) (removeLeaf (mx 3 1))                    
                    <||> when (i == 2) (removeLeaf (poly `withDegree` Just 2)))
        

trivials = 
  try toInnermost 
  >>> try irr
  >>| fastest [ dependencyPairs >>| checkTrivial
              , dependencyTuples >>| checkTrivial] 
  where 
    checkTrivial = try usableRules >>> try trivialDP >>| fastest [mx 1 0, mx 2 0]
    
trivialDP = force $ 
  te removeInapplicable
  >>> te removeHeads
  >>> try cleanSuffix 
  >>> try trivial



----------------------------------------------------------------------
--- 2012 strategies
        
--- innermost
        

directs to = 
  timeout to (te (compse 1)
              >>> te (compse 2)
              >>> te (compse 3)
              >>> te (compse 4)
              >>| empty)                           
  `orBetter` timeout (min 5 to) matchbounds
  `orBetter` timeout to ( bsearch "popstar" (spopstar `withDegree`) )
  `orBetter` timeout to popstarPS
  where compse i = withProblem $ \ prob ->
          whenNonTrivial $ 
            decomposeAnyWith (spopstarPS `withDegree` Just i) -- binary search auf grad
            <> (when (i == 2 || i == 3) (decomposeAnyWith (poly `withDegree` Just i))
                <||> wg i i
                <||> decomposeAnyWith (mx i i)
                <||> when (i < 4) (decomposeAnyWith (mx (i + 1) i)))
          
                               
shiftLeafs = removeLeafs 0 <||> (removeLeafs 1 <> removeLeafs 2)

removeFirstCongruence proc = 
  decomposeDG `solveUpperWith` (try simps >>| proc) >>> try simps


isTrivialCWDG cwdg = 
  all null $ [ DG.successors cwdg n | n <- DG.roots cwdg ]

rc2012DPi = 
  toDP >>!! te (withCWDG trans) >>! basics
  where 
    trans cwdg 
      | isTrivialCWDG cwdg = some $ shiftLeafs 
      | otherwise = some $ timeout 15 shiftLeafs <> rc
      where 
        proc = try simps >>> try simpDPRHS >>> te shiftLeafs >>! basics
        rc = removeFirstCongruence proc

rc2012RCi = 
  try irr >>! (directs 59 `orBetter` timeout 59 rc2012DPi)

    
----------------------------------------------------------------------
--- ZK setup
----------------------------------------------------------------------                


simple False = some $ 
  try toInnermost >>> try shiftRules >>! basics
  where 
    shiftRules = successive [te (shiftSemanticBy decomposeAnyWith i) | i <- [1..4]]
simple True = some $ 
  try toInnermost 
  >>> toDP >>> try shiftRulesSuccessively >>> te shiftLeaf >>! basics
  where 
    shiftLeaf = rl 1 <> rl 2 <> rl 3 <> rl 4
    rl = shiftSemanticBy removeLeaf

               
----------------------------------------------------------------------
--- ICC'12
----------------------------------------------------------------------                
                  

bspopstar = bsearch "sPOP* (PS)" (\ md -> spopstar `withDegree` md)
bspopstarPS = bsearch "sPOP* (PS)" (\ md -> spopstarPS `withDegree` md)


simpleICC usePOP = some $ 
  try toInnermost >>> try shiftRules >>| empty
  where 
    shiftRules = successive [te (shift i) | i <- [1..4]]
    shift i = whenNonTrivial $ 
              (when usePOP (decomposeAnyWith (spopstarPS `withDegree` Just i)))
              <> (decomposeAnyWith (mx i i `withUsableArgs` False)
                  <||> decomposeAnyWith (mx (i+1) i  `withUsableArgs` False)
                  <||> when (i == 2 || i == 3) (decomposeAnyWith (poly  `withUsableArgs` False `withDegree` Just i)))

----------------------------------------------------------------------
--- misc
----------------------------------------------------------------------                

checkArity = problems >>= mapM_ check
  where check p | null invalid = putStrLn "Arities are OK"
                | otherwise    = 
                  do putStrLn "Following symbols occur with miss-matching arities"
                     mapM_ (\ f -> pprint (f,sig)) invalid
        
          where sig = TR.signature p
                invalid = [ f | (f,i) <- symsWithArities
                              , (g,j) <- symsWithArities
                              , f == g , i /= j]
                symsWithArities = nub $ concatMap syms $ Set.toList $ TR.subterms $ TR.allComponents p
                nub = Set.toList . Set.fromList
                syms (TR.Var _) =  []
                syms (TR.Fun f ts) = [(f, length ts)]

